/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * decode.c: decoding loop
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <speex/speex.h>

#include "speextalk.h"

static int bandwidth_in_acc = 0;
static unsigned packet_timestamp_in = 0;
static unsigned sequence_in = 0;

/*
 * Prepares the playing channel according to the selected audio format.
 */
static void
prepare_playing_with_af(void)
{
    const audio_format_t *af;

    check_audio_format(audio_format_in, "remote ");
    af = &audio_formats[audio_format_in - 1];
    prepare_playing(af->sample_rate, af->channels);
}

/*
 * Receives a packet and stores it in the global receive buffer. Dies on
 * fatal error.
 */
void
receive_packet(void)
{
#if 0
    struct pollfd fd = { .fd = sock, .events = POLLIN };
#endif

    while(1) {
#if 0
	poll(&fd, 1, 0);
	printf("Would block: %d\n", (fd.revents & POLLIN) == 0);
#endif
	if((l_recv_buffer = recv(sock, recv_buffer,
	    sizeof(recv_buffer), 0)) < 0) {
	    if(errno == ECONNREFUSED) {
		pthread_mutex_lock(&global_mutex);
		running = 0;
		pthread_mutex_unlock(&global_mutex);
		pthread_exit(NULL);
	    }
	    if(errno == EINTR)
		continue;
	    fatal_system("recv");
	}
	update_bandwidth(&bandwidth_in, &bandwidth_in_acc,
	    &packet_timestamp_in, l_recv_buffer);
	return;
    }
}

/*
 * Waits until the first voice packet, inits the PCM. Then repeatedly
 * receives and plays packets. Exits when the running global tells to.
 * Must be called with a coherent receive buffer (size=0 is ok).
 */
void *
decode_loop(void *dummy)
{
    (void)dummy;
    SpeexBits bits;
    void *dec_state;
    int frame_size;
    unsigned char *p;
    int lp;
    int obuf_p = 0;
    int n, f;
    int yes = 1;
    unsigned seq;
    int run;

    while(1) {
	pthread_mutex_lock(&global_mutex);
	run = running;
	pthread_mutex_unlock(&global_mutex);
	if(!run)
	    break;
	if(l_recv_buffer >= 4) {
	    if(recv_buffer[0] == 1) {
		audio_format_in = recv_buffer[1];
		//printf("Receiving audio format %d\n", audio_format_in);
		break;
	    }
	}
	receive_packet();
    }
    prepare_playing_with_af();
    speex_bits_init(&bits);
    dec_state =
	speex_decoder_init(audio_formats[audio_format_in - 1].speex_mode);
    speex_decoder_ctl(dec_state, SPEEX_GET_FRAME_SIZE, &frame_size);
    speex_decoder_ctl(dec_state, SPEEX_SET_ENH, &yes);

    int obuf_size = frame_size + frame_size / 2;
    short obuf[obuf_size];

    while(1) {
	pthread_mutex_lock(&global_mutex);
	run = running;
	pthread_mutex_unlock(&global_mutex);
	if(!run)
	    break;
	if(l_recv_buffer >= 4 && recv_buffer[0] == 1) {
	    seq = (recv_buffer[2] << 8) + recv_buffer[3];
	    if(((seq - sequence_in) & 0xFFFF) > 100) {
		fprintf(stderr, "A lot of lost frames.");
		sequence_in = seq;
	    }
	    p = recv_buffer + 4;
	    lp = l_recv_buffer - 4;
	    while(lp >= 2 || seq != sequence_in) {
		if(seq != sequence_in) {
		    speex_decode_int(dec_state, NULL, obuf + obuf_p);
		} else {
		    f = ((int)p[0] << 8) + p[1];
		    p += 2;
		    lp -= 2;
		    if(lp < f) {
			fprintf(stderr, "Truncated packet.\n");
			break;
		    }
		    speex_bits_read_from(&bits, p, f);
		    p += f;
		    lp -= f;
		    speex_decode_int(dec_state, &bits, obuf + obuf_p);
		    seq++;
		}
		sequence_in++;
		sequence_in &= 0xFFFF;
		update_power(&power_in, obuf + obuf_p, frame_size);
		obuf_p += frame_size;
		while(obuf_size - obuf_p < frame_size) {
		    if((n = play_samples(obuf, obuf_p)) < 0)
			continue;
		    obuf_p -= n;
		    memmove(obuf, obuf + n, obuf_p * 2);
		}
	    }
	} else {
	    fprintf(stderr, "Ignoring invalid packet.\n");
	}
	receive_packet();
    }
    return(NULL);
}
